@@ -27,9 +27,13 @@
 */
 
 #include "config.h"
+#include <stdlib.h>
+/* That _is_ real now */
+#define ACCEPT_HEAD "Accept: audio/mpeg, audio/x-mpeg, audio/x-mpegurl, audio/x-scpls, application/pls, */*\r\n"
+char *proxyurl = NULL;
+
 #if !defined(WIN32) && !defined(GENERIC)
 
-#include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 
@@ -56,35 +60,34 @@
 #include "config.h"
 #include "mpg123.h"
 #include "debug.h"
+#include "stringbuf.h"
+#include "icy.h"
 
 #ifndef INADDR_NONE
 #define INADDR_NONE 0xffffffff
 #endif
 
-void writestring (int fd, char *string)
+int writestring (int fd, char *string)
 {
 	int result, bytes = strlen(string);
 
 	while (bytes) {
 		if ((result = write(fd, string, bytes)) < 0 && errno != EINTR) {
-			perror ("write");
-			exit (1);
+			perror ("writing http string");
+			return 0;
 		}
 		else if (result == 0) {
-			fprintf (stderr, "write: %s\n",
-				"socket closed unexpectedly");
-			exit (1);
+			error("write: socket closed unexpectedly");
+			return 0;
 		}
 		string += result;
 		bytes -= result;
 	}
+	return 1;
 }
 
 int readstring (char *string, int maxlen, FILE *f)
 {
-#if 0
-	char *result;
-#endif
 	int pos = 0;
 
 	while(pos < maxlen) {
@@ -95,25 +98,16 @@ int readstring (char *string, int maxlen, FILE *f)
 			}
 		}
 		else if(errno != EINTR) {
-			fprintf (stderr, "Error reading from socket or unexpected EOF.\n");
-			exit(1);
+			error("Error reading from socket or unexpected EOF.");
+			string[pos] = 0;
+			/* bail out to prevent endless loop */
+			return -1;
 		}
 	}
 
 	string[pos] = 0;
 
 	return pos;
-
-#if 0
-	do {
-		result = fgets(string, maxlen, f);
-	} while (!result  && errno == EINTR);
-	if (!result) {
-		fprintf (stderr, "Error reading from socket or unexpected EOF.\n");
-		exit (1);
-	}
-#endif
-
 }
 
 void encode64 (char *source,char *destination)
@@ -218,16 +212,39 @@ char *url2hostport (char *url, char **hname, unsigned long *hip, unsigned int *p
 	return (cptr);
 }
 
-char *proxyurl = NULL;
+char* get_header_val(const char *hname, char* response, size_t *length)
+{
+	char *tmp = NULL;
+	size_t len = 0;
+	size_t prelen = strlen(hname);
+	/* if header name found, next char is at least something, so just check for : */
+	if(!strncasecmp(hname, response, prelen) && (response[prelen] == ':'))
+	{
+		++prelen;
+		if((tmp = strchr(response, '\r')) != NULL ) tmp[0] = 0;
+		if((tmp = strchr(response, '\n')) != NULL ) tmp[0] = 0;
+		len = strlen(response)-prelen;
+		tmp = response+prelen;
+		while(len && ((tmp[0] == ' ') || (tmp[0] == '\t')))
+		{
+			++tmp;
+			--len;
+		}
+	}
+	*length = len;
+	return tmp;
+}
+
 unsigned long proxyip = 0;
 unsigned int proxyport;
 
-/* That is not real ... I should really check the type of what I get! */
-#define ACCEPT_HEAD "Accept: audio/mpeg, audio/x-mpegurl, audio/x-scpls, */*\r\n"
 /* needed for HTTP/1.1 non-pipelining mode */
 /* #define CONN_HEAD "Connection: close\r\n" */
 #define CONN_HEAD ""
 
+/* shoutcsast meta data: 1=on, 0=off */
+#define ACCEPT_ICY_META "Icy-MetaData: 1\r\n"
+
 char *httpauth = NULL;
 char *httpauth1 = NULL;
 
@@ -244,6 +261,7 @@ int http_open (char* url, char** content_type)
 	unsigned int myport = 0;
 	int sock;
 	int relocate, numrelocs = 0;
+	int ret = 0; /* return value from readstring */
 	struct sockaddr_in server;
 	FILE *myfile;
 	/*
@@ -286,7 +304,8 @@ int http_open (char* url, char** content_type)
 	purl = (char *)malloc(purl_size);
 	if (!purl) {
 		fprintf (stderr, "malloc() failed, out of memory.\n");
-		exit (1);
+		sock = -1;
+		goto exit;
 	}
 	/* make _sure_ that it is null-terminated */
 	purl[purl_size-1] = 0;
@@ -317,8 +336,9 @@ int http_open (char* url, char** content_type)
 	if(httpauth1 != NULL) free(httpauth1);
 	httpauth1 = (char *)malloc((strlen(purl) + 1));
 	if(!httpauth1) {
-		fprintf(stderr, "malloc() failed, out of memory.\n");
-		exit(1);
+		error("malloc() failed, out of memory.");
+		sock = -1;
+		goto exit;
 	}
         if (getauthfromURL(purl,httpauth1,strlen(purl)) < 0) {
 		sock = -1;
@@ -331,9 +351,10 @@ int http_open (char* url, char** content_type)
 	 * ACCEPT_HEAD               strlen(ACCEPT_HEAD)
 	 * "Authorization: Basic \r\n"	23
 	 * "\r\n"			 2
+	 * ... plus the other predefined header lines
 	 */
 	linelengthbase = 62 + strlen(PACKAGE_NAME) + strlen(PACKAGE_VERSION)
-	                 + strlen(ACCEPT_HEAD) + strlen(CONN_HEAD);
+	                 + strlen(ACCEPT_HEAD) + strlen(CONN_HEAD) + strlen(ACCEPT_ICY_META);
 
 	if(httpauth) {
 		tmp = (strlen(httpauth) + 1) * 4;
@@ -369,8 +390,9 @@ int http_open (char* url, char** content_type)
 			if(request_url != NULL)	request_url[request_url_size-1] = '\0';
 			else
 			{
-				fprintf(stderr, "malloc() failed, out of memory.\n");
-				exit(1);
+				error("malloc() failed, out of memory.");
+				sock = -1;
+				goto exit;
 			}
 		}
 		/* used to be url here... seemed wrong to me (when loop advanced...) */
@@ -411,8 +433,9 @@ int http_open (char* url, char** content_type)
 			response = (char *)malloc((linelength + 1));
 
 			if ((request == NULL) || (response == NULL)) {
-				fprintf (stderr, "malloc() failed, out of memory.\n");
-				exit(1);
+				error("malloc() failed, out of memory.");
+				sock = -1;
+				goto exit;
 			}
 
 			strcpy (request, "GET ");
@@ -461,8 +484,9 @@ int http_open (char* url, char** content_type)
 			response = (char *)malloc((linelength + 1));
 
 			if ((request == NULL) || (response == NULL)) {
-				fprintf (stderr, "malloc() failed, out of memory.\n");
-				exit(1);
+				error("malloc() failed, out of memory.");
+				sock = -1;
+				goto exit;
 			}
 			
 			strcpy (request, "GET ");
@@ -490,28 +514,28 @@ int http_open (char* url, char** content_type)
 		} */
 		strcat (request, ACCEPT_HEAD);
 		strcat (request, CONN_HEAD);
+		strcat (request, ACCEPT_ICY_META);
 		server.sin_family = AF_INET;
 		server.sin_port = htons(myport);
 		server.sin_addr.s_addr = myip;
+#define http_failure close(sock); sock=-1; goto exit;
 		if ((sock = socket(PF_INET, SOCK_STREAM, 6)) < 0) {
 			perror ("socket");
-                       goto exit;
+			sock=-1;
+			goto exit;
 		}
 		if (connect(sock, (struct sockaddr *)&server, sizeof(server))) {
 			perror ("connect");
-                       close(sock);
-                       sock = -1;
-                       goto exit;
+			http_failure;
 		}
-
 		if (httpauth1 || httpauth) {
 			char *buf;
 			strcat (request,"Authorization: Basic ");
 			if(httpauth1) {
 				buf=(char *)malloc((strlen(httpauth1) + 1) * 4);
 				if(!buf) {
-					fprintf(stderr, "Error allocating sufficient memory for http authentication.  Exiting.");
-					exit(1);
+					error("malloc() failed for http auth, out of memory.");
+					http_failure;
 				}
 				encode64(httpauth1,buf);
 				free(httpauth1);
@@ -519,8 +543,8 @@ int http_open (char* url, char** content_type)
 			} else {
 				buf=(char *)malloc((strlen(httpauth) + 1) * 4);
 				if(!buf) {
-					fprintf(stderr, "Error allocating sufficient memory for http authentication.  Exiting.");
-					exit(1);
+					error("malloc() for http auth failed, out of memory.");
+					http_failure;
 				}
 				encode64(httpauth,buf);
 			}
@@ -532,22 +556,26 @@ int http_open (char* url, char** content_type)
 		strcat (request, "\r\n");
 		
 		debug1("<request>\n%s</request>",request);
-		writestring (sock, request);
+		if(!writestring (sock, request)){ sock = -1; goto exit; }
 		if (!(myfile = fdopen(sock, "rb"))) {
 			perror ("fdopen");
-                       close(sock);
-                       sock = -1;
-                       goto exit;
+			http_failure;
 		}
 		relocate = FALSE;
 		purl[0] = '\0';
-		if (readstring (response, linelength-1, myfile)
-		    == linelength-1) {
-			fprintf(stderr, "Command exceeds max. length\n");
-			close(sock);
-			sock = -1;
-			goto exit;
+		#define safe_readstring \
+		ret = readstring(response, linelength-1, myfile); \
+		if(ret == linelength-1) \
+		{ \
+			error("HTTP response line exceeds max. length"); \
+			http_failure; \
+		} \
+		else if(ret < 0) \
+		{ \
+			error("readstring failed"); \
+			http_failure; \
 		}
+		safe_readstring;
 		debug1("<response>\n%s</response>",response);
 		if ((sptr = strchr(response, ' '))) {
 			switch (sptr[1]) {
@@ -556,27 +584,21 @@ int http_open (char* url, char** content_type)
 				case '2':
 					break;
 				default:
-                                       fprintf (stderr,
-                                                "HTTP request failed: %s",
-                                                sptr+1); /* '\n' is included */
-                                       close(sock);
-                                       sock = -1;
-                                       goto exit;
+					fprintf (stderr, "HTTP request failed: %s", sptr+1); /* '\n' is included */
+					http_failure;
 			}
 		}
 		do {
-			if (readstring (response, linelength-1, myfile)
-			    == linelength-1) {
-				fprintf(stderr, "URL exceeds max. length\n");
-				close(sock);
-				sock = -1;
-				goto exit;
-			}
+			safe_readstring;
 			if (!strncmp(response, "Location: ", 10))
 			{
 				size_t needed_length;
 				char* prefix = (char*) malloc(strlen(request_url)+1);
-				if(prefix == NULL){ error("out of memory here... cannot handle that gracefully (yet)"); exit(1); }
+				if(prefix == NULL)
+				{
+					error("malloc() failed, out of memory.");
+					http_failure;
+				}
 				strcpy(prefix, request_url); /* fits and is terminated! */
 				
 				debug1("request_url:%s", request_url);
@@ -618,9 +640,7 @@ int http_open (char* url, char** content_type)
 					purl = realloc(purl, purl_size);
 					if(purl == NULL)
 					{
-						close(sock);
-						sock = -1;
-						goto exit;
+						http_failure;
 					}
 					/* Why am I always so picky about the trailing zero, nobody else seems to care? */
 					purl[purl_size-1] = 0;
@@ -639,24 +659,14 @@ int http_open (char* url, char** content_type)
 			}
 			else
 			{
+				char *tmp;
+				size_t len;
 				/* watch out for content type */
-				debug1("searching for content-type... %s", response);
-				if(!strncasecmp("content-type:", response, 13))
+				debug1("searching for header values... %s", response);
+				if((tmp = get_header_val("content-type", response, &len)))
 				{
 					if(content_type != NULL)
 					{
-						char *tmp = NULL;
-						size_t len = 0;
-						
-						if((tmp = strchr(response, '\r')) != NULL ) tmp[0] = 0;
-						if((tmp = strchr(response, '\n')) != NULL ) tmp[0] = 0;
-						len = strlen(response)-13;
-						tmp = response+13;
-						while(len && ((tmp[0] == ' ') || (tmp[0] == '\t')))
-						{
-							++tmp;
-							--len;
-						}
 						if(len)
 						{
 							if(*content_type != NULL) free(*content_type);
@@ -667,17 +677,34 @@ int http_open (char* url, char** content_type)
 								(*content_type)[len] = 0;
 								debug1("got type %s", *content_type);
 							}
-							else fprintf(stderr, "Error: canno allocate memory for content type!\n");
+							else error("cannot allocate memory for content type!");
 						}
 					}
 				}
+				/* watch out for icy-name */
+				else if((tmp = get_header_val("icy-name", response, &len)))
+				{
+					if(set_stringbuf(&icy.name, tmp)) debug1("got icy-name %s", icy.name.p)
+					else error1("unable to set icy name to %s!", tmp)
+				}
+				/* watch out for icy-url */
+				else if((tmp = get_header_val("icy-url", response, &len)))
+				{
+					if(set_stringbuf(&icy.url, tmp)) debug1("got icy-url %s", icy.name.p)
+					else error1("unable to set icy url to %s!", tmp)
+				}
+				/* watch out for icy-metaint */
+				else if((tmp = get_header_val("icy-metaint", response, &len)))
+				{
+					icy.interval = atoi(tmp);
+					debug1("got icy-metaint %li", (long int)icy.interval);
+				}
 			}
 		} while (response[0] != '\r' && response[0] != '\n');
 	} while (relocate && purl[0] && numrelocs++ < HTTP_MAX_RELOCATIONS);
 	if (relocate) {
 		fprintf (stderr, "Too many HTTP relocations.\n");
-               close(sock);
-               sock = -1;
+		http_failure;
 	}
 exit:
 	if(host != NULL) free(host);
@@ -689,36 +716,11 @@ exit:
 }
 
 #else /* defined(WIN32) || defined(GENERIC) */
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
 
-extern int errno;
-
-#include "mpg123.h"
-
-/* stubs for Win32 */
-
-void writestring (int fd, char *string)
-{
-}
-
-int readstring (char *string, int maxlen, FILE *f)
-{
-}
-
-char *url2hostport (char *url, char **hname, unsigned long *hip, unsigned int *port)
-{
-}
-
-char *proxyurl = NULL;
-unsigned long proxyip = 0;
-unsigned int proxyport;
-
-#define ACCEPT_HEAD "Accept: audio/mpeg, audio/x-mpegurl, */*\r\n"
-
-int http_open (char *url)
+/* stub */
+int http_open (char* url, char** content_type)
 {
+	return -1;
 }
 #endif
 
